﻿using UnityEngine;
using System.Collections.Generic;

/***********************

Owen Canavan

Pac-Man
©1980 Midway

Rather than 'take inspiration' from a classic 2D game, I decided to take a 2D game and try and
recreate it as closely as possible, so as not to hide failures in implementation behind 'game design'.
I chose Pac-Man due to the extensive breakdown descriptions of the operation logic contained 
in "The Pac-Man Dossier", compiled by Jamey Pittman which can be found at 
http://home.comcast.net/~jpittman2/pacman/pacmandossier.html
While the logic of the game is described it is from a players perspective, not a developers, and
as such does not provide insight on HOW to implement the features. I saw this as an opportunity 
to try and implement my own version.

Secondly, following the brief of making as much of the game in code as possible, I decided to exclude
any non-code assets (audio, visual) in the project. All textures used in the game are made at runtime
within the scripts. The entire project consists of 11 C# scripts and one custom simple unlit tranparency 
fragment shader.

BUGS:
*If the game switches to FRIGHTENED mode while a ghost is leaving the ghost house, they will not return 
to their full speed for the level on exiting FRIGHTENED mode 

MISSING CONTENT:
*Bonus Fruit
*Cutscene animations
*Pseudo-Random pathing for ghosts in FRIGHTENED mode
*Ghosts flashing when exiting FRIGHTENED mode
***********************/


//This is the main manager script for the game and the only MonoBehaviour derived script.
//This script is attached to the Main Camera GameObject in the scene, which is the only
//object that exists in the scene before runtime.

//game states
public enum GamePhase {MENU, STARTUP, NEWLEVEL, READY, GAME, KILL, DYING, DEAD, GAMEOVER, COMPLETE, FLASH};

//Main game manager class
public class GameManager : MonoBehaviour {

	static GameManager _instance; //singleton
	Menu menu;

	public Map map; //contains the map and objects such as dots and pills

	public float vert = 0f;
	public float horz = 0f;
	public bool inputLeft = false;
	public bool inputRight = false;
	public bool inputUp = false;
	public bool inputDown = false;

	//Entities
	Pacman _pacman;
	
	Blinky _blinky;
	Pinky _pinky;
	Inky _inky;
	Clyde _clyde;

	//for tracking runtime states
	GamePhase phase;

	Ghost[] ghosts;
	Ghost[] house_ghosts; //ghosts current;y in the ghost house
	GhostMode ghostMode = GhostMode.SCATTER; //global ghost mode
	GhostMode lastMode; //last ghost mode
	Ghost eaten_ghost = null; //last ghost eaten

	//time in seconds for duration of FRIGHTENED mode for level 1-20 & 21+
	int[] fright_times = new int[21]{6,5,4,3,2,5,2,2,1,5,2,1,1,3,1,1,0,1,0,0,0};
	
	
	//SCATTER,CHASE,SCATTER, etc... times for level 1, levels 2-4, and level 5 or greater
	float[,] mode_times = new float[3,8]{
		{7f,20f,7f,20f,5f,20f,5f,Mathf.Infinity},
		{7f,20f,7f,20f,5f,1033f,1/30f,Mathf.Infinity},
		{5f,20f,5f,20f,5f,1037f,1/30f,Mathf.Infinity}
	};
	float mode_timer; //for tracking mode_times
	int mode_index = 0; //index position in the current mode_times array
	float fright_timer;
	int pacman_pause_frames = 0; //pacman pauses for 1 frame after eating a dot, and 3 frames after eating a pill

	int _level = 0;
	int _dots = 244; //total dots+pills in a level
	int lives = 3;
	int global_dot_counter = 0; //used to contorl when ghosts leave the Ghost House, when in GLOBAL house mode
	enum HouseMode { GHOSTS, GLOBAL }; //the modes of ghost house control
	HouseMode house_mode = HouseMode.GHOSTS; //start in GHOSTS mode
	float dot_timer = 0f; //additional ghost house control to avoid ghosts being stuck in ghost house

	//UI text
	Word player1;
	Word score_label;
	Word hiscore_label;
	Word score;
	Word hiscore;
	Word ready;
	Word gameover;
	Word[] words;

	Bonus bonus; //appears when ghosts eaten
	int ghosts_eaten = 0; //total ghosts eaten this FRIGHTENED mode

	int score_value; //current score value
	int hiscore_value;
	int score_last_frame; //used to determine when to update socre font frames
	int hiscore_last_frame;

	bool paused = false; //used to halt update() on ghosts and pacman
	float phase_timer = 0;
	float game_timer = 0;
	int flashes = 0; //level complete flash counter
	
	
	//A pseudo-singleton (it doesn't ensure only a single instance of GameManager). 
	//Allows global access of GameManager object instance without having to share this with every other class
	public static GameManager Instance
	{
		get { return _instance; }
	}

	void Awake(){
		_instance = this; //store the instance
		Application.targetFrameRate = 60; //try and set the framerate, doesn't necessairly work in editor
	}

	void Start () {
		Debug.Log("Welcome to Pac-Man.\nPress 1 to insert a coin");

		//adjust camera size and position so the map width fills entire screen
		camera.orthographicSize = 18f;
		transform.position = new Vector3(14f,-14.5f,-10f);

		//create a map
		map = new Map();
		//this has build the map and populated the objects, but we don't want to see it yet so we hide it
		map.enabled = false;

		//we actually want to start in a menu
		menu = new Menu();

		//build UI that is always displayed at top of the screen
		score_label = new Word("1UP", new Vector3(3f, 3f, -1f), Color.white);
		hiscore_label = new Word("HIGH SCORE", new Vector3(9f, 3f, -1f), Color.white);
		score = new Word("     00", new Vector3(0f, 2f, -1f), Color.white);
		hiscore = new Word("     00", new Vector3(10f, 2f, -1f), Color.white);
		hiscore.enabled = false;


		lastMode = ghostMode;
		//create the game Entities
		_pacman = new Pacman();

		_blinky = new Blinky(false);
		_pinky = new Pinky(false);
		_inky = new Inky(false);
		_clyde = new Clyde(false);
		ghosts = new Ghost[]{_blinky, _pinky, _inky, _clyde};
		house_ghosts = new Ghost[3]; //prepare Pinky, Inky and Clyde's home. Blinky is a drifter

		//in game on map text
		player1 = new Word("PLAYER ONE", new Vector3(9.5f,-11.5f, -1f), new Color(0f,1f,1f));
		player1.enabled = false;
		ready = new Word("READY!", new Vector3(12f, -17.5f, -1f), Color.yellow);
		ready.enabled = false;
		gameover = new Word("GAME  OVER", new Vector3(9.5f, -17.5f, -1f), Color.red);
		gameover.enabled = false;

		//score bonus ui text
		bonus = new Bonus();

		//store for easy update.
		// this was intended up flashing words.
		words = new Word[]{player1, score_label, hiscore_label, hiscore, score, hiscore, ready};

		//start in the menu
		changePhase(GamePhase.MENU);
	}

	//menu/intro is complete, start the game already!
	public void StartFromMenu(){
		changePhase(GamePhase.STARTUP);
	}

	public void gameOver(){
		map.enabled = false; //hide the map and objects
		gameover.enabled = false; //hide the game over message
		changePhase(GamePhase.MENU); //go back to the menu
	}

	//reset values for a new level, prepare entities
	public void startLevel(int l){
		_level = l;
		_dots = 244;
		house_mode = HouseMode.GHOSTS;
		if(pacman!=null){
			pacman.startLevel();
		}
		foreach(Ghost ghost in ghosts){
			if(ghost != null)
				ghost.startLevel();
		}
	}

	//pacman dies, reset the entities, but don't fully reset the level
	public void resetLevel(){
		if(pacman!=null){
			pacman.firsPositions();
		}
		foreach(Ghost ghost in ghosts){
			if(ghost != null)
				ghost.firsPositions();
		}
	}

	protected void changePhase(GamePhase p){
		switch(p){
			case GamePhase.MENU : 
				phase = p;
				menu.start(); //start the menu 'animation' sequence
				break;
			case GamePhase.STARTUP : //set once at beginning of a new game
				paused = true;
				//enable ui labels
				player1.enabled = true;
				ready.enabled = true;
				map.enabled = true;
				phase_timer = 2f; //interval until next pahse change
				score_value = 0;
				score.text("     00"); //reset score
				addToScore(0);
				lives = 3;
				_level = 0;
				map.LevelReady(); //repopulate the map
				break;
			case GamePhase.NEWLEVEL:
				map.LevelReady(); //repopulate map objects after completed level
				p = GamePhase.READY;
				goto case GamePhase.READY; //jump immediately to READY
			case GamePhase.READY :
				paused = true;
				ready.enabled = true;
				player1.enabled = false;
				phase_timer = 2f;	
				mode_index = 0; //resetLevel SCATTER,CHASE cycle
				mode_timer = 0f;
				if(phase == GamePhase.DEAD){ //partial reset or full reset
					resetLevel(); 
				} else {
					startLevel(_level+1);
				}
				ghostMode = GhostMode.SCATTER;
				break;
			case GamePhase.GAME : //this is the main game mode
				ready.enabled  = false ;
				paused = false;
				phase_timer = 0;
				hiscore.enabled = true;
				break;
			case GamePhase.KILL: //pacman killed, temp pause before reset
				pacman.setMode(Pacman.Modes.KILL);
				foreach(Ghost ghost in ghosts){
					if(ghost != null){
						ghost.halt();
					}
				}
				phase_timer = 1f;
				lives--;
				break;
			case GamePhase.DYING: //pacman death animation is playing
				pacman.setMode(Pacman.Modes.DYING);
				foreach(Ghost ghost in ghosts){
					if(ghost != null)
						ghost.enabled = false;
				}
				phase_timer = 2;
				break;
			case GamePhase.DEAD: //pacman death animation is complete
				phase_timer = 1f;
				break;
			case GamePhase.GAMEOVER: //short display of GAME OVER before going back to menu
				gameover.enabled = true;
				pacman.enabled = false;
				phase_timer = 2f;
				break;
			case GamePhase.COMPLETE: //level cleared
				paused = true;
				phase_timer = 2f;
				flashes = 9;
				break;
			case GamePhase.FLASH: //flash level 4 times after cleared
				if(flashes == 9){
					foreach(Ghost ghost in ghosts){
						if(ghost != null)
							ghost.enabled = false;
					}
				}
				if(--flashes > 0){
					phase_timer = 0.2f;
					map.enabled = !map.enabled;
				} else {
					goto case GamePhase.NEWLEVEL;
				}
				break;
		}
		phase = p;
	}

	// Update is called once per frame
	void FixedUpdate () {
		if(phase == GamePhase.MENU){ //update menu
			menu.update(Time.deltaTime);
		} else { //update game
			if(phase_timer > 0){ //check for phase changes
				phase_timer -= Time.deltaTime;
				if(phase_timer <= 0){
					if(phase == GamePhase.STARTUP){
						changePhase(GamePhase.READY);
					} else if( phase == GamePhase.READY){
						changePhase(GamePhase.GAME);
					} else if (phase == GamePhase.KILL){
						changePhase(GamePhase.DYING);
					} else if (phase == GamePhase.DYING){
						if(lives > 0){
							changePhase(GamePhase.DEAD);
						} else {
							changePhase(GamePhase.GAMEOVER);
						}
					} else if (phase == GamePhase.DEAD){
						changePhase(GamePhase.READY);
					} else if (phase == GamePhase.GAMEOVER){
						gameOver();
					} else if (phase == GamePhase.COMPLETE){
						changePhase(GamePhase.FLASH);
					} else if (phase == GamePhase.FLASH){
						changePhase(GamePhase.FLASH);
					} else {
						paused = false;
						phase_timer = 0;
					}
				}
			}
			if(game_timer > 0){ //check game pauses
				game_timer -= Time.deltaTime;
				if(game_timer <= 0){
					paused = false;
					game_timer = 0;
					if(eaten_ghost != null){
						eaten_ghost.enabled = true; //reshow eaten ghost, now eyes after bonus text pause
						pacman.enabled = true; //reshow pacman after same
						bonus.enabled = false; //hide bonus text
						eaten_ghost = null;
					}
				}
			}
			if(!paused) {
				if(ghostMode == GhostMode.FRIGHTENED){ //countdown FRIGHTENED duration
					fright_timer += Time.deltaTime;
					if(fright_timer > fright_times[levelIndex]){
						changeMode(lastMode);
					}
				} else if (ghostMode == GhostMode.CHASE || ghostMode == GhostMode.SCATTER){ //check of SCATTER/CHASE cycle change
					mode_timer += Time.deltaTime;
					int level_index = (level == 1 ? 0 : (level >= 2 && level <= 4 ? 1 : 2));
					if(mode_timer > mode_times[level_index, mode_index]){
						mode_index++;
						mode_timer = 0f;
						GhostMode m = ghostMode == GhostMode.SCATTER ? GhostMode.CHASE : GhostMode.SCATTER;
						changeMode(m);
					}
				}
				if(pacman_pause_frames > 0){ //pacman pause after eating dot (1 frame) or pill (3 frame)
					pacman_pause_frames--;
				} else {
					pacman.update(Time.deltaTime); //otherwise update pacman
				}
				foreach(Ghost ghost in ghosts){ //update ghosts
					if(ghost != null)
						ghost.update(Time.deltaTime);
				}

				dot_timer += Time.deltaTime; //add to time since dot/pill eaten

				if(dot_timer > (level < 5 ? 4f : 3f)){ //been a while since somthign eaten, eject a ghost is there is one at home
					ejectGhost();
				}
			}
		
			foreach(Word word in words){ //update words. unused
				if(word != null){
					word.update(Time.deltaTime);
				}
			}

			//update score and hiscore if changed since last frame
			if(score_last_frame != score_value){
				score.text(score_value);
				score_last_frame = score_value;
			}
			if(hiscore_last_frame != hiscore_value){
				hiscore.text(hiscore_value);
				hiscore_last_frame = hiscore_value;
			}
		}
	}

	void Update(){
		//manual pause
		if(Input.GetKeyDown(KeyCode.P)){
			paused = !paused;
		}
		
		horz = Input.GetAxis("Horizontal");
		vert = Input.GetAxis("Vertical");
		
		inputLeft = Input.GetButton("LEFT");
		inputRight = Input.GetButton("RIGHT");
		inputUp = Input.GetButton("UP");
		inputDown = Input.GetButton("DOWN");
	}

	//add value to score. update hiscore to match if high
	protected void addToScore(int v){
		score_value += v;
		if(score_value > hiscore_value){
			hiscore_value = score_value;
		}
	}

	//a ghost caught pacman
	public void pacmanEaten(Ghost g){
		changePhase(GamePhase.KILL);
		global_dot_counter = 0;
		house_mode = HouseMode.GLOBAL;
	}

	//pacman caught a ghost
	public void ghostEaten(Ghost ghost){
		paused = true;
		game_timer = 1f;
		ghost.mode = GhostMode.EYES1;
		ghost.enabled = false; //hide eaten ghost and pacman and pause for 1 second, and show bonus score
		pacman.enabled = false;
		eaten_ghost = ghost;
		ghosts_eaten++;
		int b = 0;
		switch(ghosts_eaten){
			case 1: b = 200; break;
			case 2: b = 400; break;
			case 3: b = 800; break;
			case 4: b = 1600; break;
			default:b = 1600; break;
		}
		addToScore(b);
		bonus.text(b, ghost.rawPos, Color.cyan);
	}

	//map says a dot was eaten
	public void dotEaten(){
		addToScore(10);
		pacman_pause_frames += 1;
		dots--;
	}

	//map says pill was eaten. enter FRIGHTENED mode if able this level
	public void pillEaten(){
		addToScore(50);
		pacman_pause_frames += 3;
		dots--;
		fright();
	}
	
	//getter/setter for total dots. if none left, level is complete
	public int dots{
		get { return _dots; }
		protected set { 
			_dots = value;
			if(_dots == 0){
				//level complete
				changePhase(GamePhase.COMPLETE);
			} else {
				addGhostDot();
			}
		}
	}

	//a dot was eaten, see if this lets a ghost out of the ghost house
	protected void addGhostDot(){
		dot_timer = 0;
		if(house_mode == HouseMode.GHOSTS){
			//this is the primary mode of checking to left ghosts out. levels start this way
			//only one ghost can be processed at a time, in the order of Pinky, Inky then Clyde. Poor Clyde, always picked last.
			//In this mode Pinky alwyas leave immediately, as do the other two after a couple levels.
			for(int i = 0; i < 3; i++){
				if(house_ghosts[i] != null){
					bool leaving = (house_ghosts[i] as Ghost).dotEaten();
					if(leaving){
						house_ghosts[i] = null;
					}
					break;
				}
			}
		} else if (house_mode == HouseMode.GLOBAL){
			//if the level resets after pacman losing a life, the game changes to this global check
			//now ghosts can only get out is a specific global value is hit.
			//this mode can only be stopped if Clyde is in the house when the counter reaches 32. this counter does not reset otherwise
			//to avoid ghosts getting stuck due to this, the timer since a dot was last eaten can also eject a ghost
			global_dot_counter++;
			if(global_dot_counter == 7){
				if(house_ghosts[0] != null){
					//Pinky is in da house
					(house_ghosts[0] as Ghost).LeaveGhostHouse();
					house_ghosts[0] = null;
				}
			} else if(global_dot_counter == 17){
				if(house_ghosts[1] != null){
					//Pinky is in da house
					(house_ghosts[1] as Ghost).LeaveGhostHouse();
					house_ghosts[1] = null;
				}
			} else if(global_dot_counter == 32){
				if(house_ghosts[2] != null){
					//Pinky is in da house
					house_mode = HouseMode.GHOSTS;
				}
			}
		}
	}

	//find a ghost to leave the ghost house, in order of Pinky, Inky, then Clyde. Blinky never stays in the ghost house
	protected void ejectGhost(){
		dot_timer = 0;
		if(house_ghosts[0] != null){
			//Pinky is in da house
			(house_ghosts[0] as Ghost).LeaveGhostHouse();
			house_ghosts[0] = null;
			return;
		}
		if(house_ghosts[1] != null){
			//Inky is in da house
			(house_ghosts[1] as Ghost).LeaveGhostHouse();
			house_ghosts[1] = null;
			return;
		}
		if(house_ghosts[2] != null){
			//Cylde is in da house
			(house_ghosts[2] as Ghost).LeaveGhostHouse();
			house_ghosts[2] = null;
			return;
		}
	}

	//a ghost tells us it has entered the hosue
	public void GhostInDaHouse(Ghost g){
		//Debug.Log(g+" in da house!");
		if(g is Pinky)
			house_ghosts[0] = g;
		else if(g is Inky)
			house_ghosts[1] = g;
		else if( g is Clyde)
			house_ghosts[2] = g;
	}

	//enter FRIGHTENED mode, if able
	protected void fright(){
		if(fright_times[levelIndex] > 0){
			ghosts_eaten = 0;
			fright_timer = 0;
			if(ghostMode != GhostMode.FRIGHTENED)
				changeMode(GhostMode.FRIGHTENED);
		}
	}

	//global change of ghost mode, mainly either my a pill being eaten or the SCATTER/CHASE cycle
	void changeMode(GhostMode m){
		lastMode = ghostMode;
		ghostMode = m;
		foreach(Ghost ghost in ghosts){
			if(ghost != null && ghost.mode != GhostMode.EYES1 && ghost.mode != GhostMode.EYES2){
				ghost.mode = ghostMode;
			}
		}
	}

	//some public accessors

	public Pacman pacman {
		get { return _pacman;}
	}

	public Blinky blinky {
		get { return _blinky; }
	}

	public int level {
		get { return _level == 0 ? 1 : _level; }
	}

	//level array values only go to 21 slots. values don't change after that the 21st level
	public int levelIndex {
		get { return level > 21 ? 20 : level-1; }
	}

	public GhostMode CurrentGhostMode {
		get { return ghostMode; }
	}

	public GhostMode LastGhostMode {
		get { return lastMode; }
	}
}
