using UnityEngine;
using System.Collections.Generic;
using System;

public enum GhostMode { NONE, CHASE, SCATTER, FRIGHTENED, GHOSTHOUSE, LEAVING, OUTSIDE, EYES1, EYES2};

public class Ghost : Entity {

	//an array to define the pixels for every required frame (2 movement frames from 3 directions plus frightened.
	protected int[,] pixelData = new int[SIZE,SIZE]{
		{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
		{ 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0},
		{ 0, 0, 0, 0,14,14, 1, 1, 1, 1,14,14, 0, 0, 0, 0},
		{ 0, 0, 0,13,14,14,13, 1, 1,13,14,14,13, 0, 0, 0},
		{ 0, 0, 1, 5, 5,13,13, 1, 1, 5, 5,13,13, 1, 0, 0},
		{ 0, 0, 7, 5, 9,15,16, 1, 7,19,15, 9,13, 1, 0, 0},
		{ 0, 0, 6, 8, 9,15,17, 1, 6,20,15, 9,10, 1, 0, 0},
		{ 0, 1, 6, 8, 4, 4,10, 1, 6, 8, 4, 4,10, 1, 1, 0},
		{ 0, 1, 1, 4,11,12,10, 1, 1, 4,11,12,10, 1, 1, 0},
		{ 0, 1, 1, 1,12,12, 1, 1, 1, 1,12,12, 1, 1, 1, 0},
		{ 0, 1, 1,18,18, 1, 1,18,18, 1, 1,18,18, 1, 1, 0},
		{ 0, 1,18, 1, 1,18,18, 1, 1,18,18, 1, 1,18, 1, 0},
		{ 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0},
		{ 0, 1, 1, 3, 1, 2, 1, 3, 3, 1, 2, 1, 3, 1, 1, 0},
		{ 0, 2, 3, 3, 0, 2, 2, 3, 3, 2, 2, 0, 3, 3, 2, 0},
		{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}
	};

	//colour of this ghosts body. redefined in each ghost type.
	protected Color primaryColor = new Color(222/255f,222/255f,222/255f);
	protected Color eyeWhite = new Color(222/255f,222/255f,222/255f);
	protected Color eyeBlue = new Color(33/255f,33/255f,222/255f);
	protected Color frightBlue = new Color(0,120/255f,248/255f);
	protected Color frightEye = new Color(216/255f,184/255f,248/255f);
	
	//the location to attempt to move towards when in SCATTER mode.
	//redefined in each ghost type.
	//this point will be outside the map near one of the four corners so that the ghost will move in a loop
	//near that corner for the duration of SCATTER mode.
	public Vector2 homePos = new Vector3(0,0);
	//target just above the ghost house in order to find way back to ghost house after been eaten (EYES1 mode)
	protected Vector2 house_target = new Vector2(12,11);
	//state tracker for entering and exiting the ghost house
	protected int house_stage = 0;
	
	//all of the animation frame sprites that have been created for this ghost. 
	//Not cached in a static as colours are different. maybe should have been cached on each ghost sub class (menu ghosts etc)
	protected Dictionary<GhostMode, Dictionary<Dir,Texture2D[]>> frames = new Dictionary<GhostMode, Dictionary<Dir, Texture2D[]>>();

	//The main ghost game modes are SCATTER, CHASE and FRIGHTENED. adding extra states to this type may have been needlessly complicate things
	//additional modes probably should have been split off.
	protected GhostMode _mode = GhostMode.NONE;
	protected GhostMode start_mode = GhostMode.NONE;
	//if in the ghos house (in or exiting) and frightened. an example of the above confusion
	protected GhostMode anim_mode_override = GhostMode.NONE;
	//when ghost game modes change, ghosts should reverse direction on entering the next tile
	protected bool doReverse = false;
	//if in the ghost house and next in line to exit, track dots eaten here is not in GLOBAL mode
	protected int dot_counter = 0;
	protected float house_speed = 0.75f;
	protected float exit_speed = 0.3f;
	protected float eyes_speed = 2f;
	//don't reemerge from the ghost house in to frightened mode after reviving
	protected bool was_eyes = false;

	//number of warning flashes on exiting frightened more for each level. currently not implementated
	protected int[] fright_flashes = new int[21]{5,5,5,5,5,5,5,5,3,5,5,3,3,5,3,3,0,3,0,0,0};
	//speeds while in FRIGHTENED more for each level
	protected float[] fright_speeds = new float[21]{
		0.5f,0.55f,0.55f,0.55f,0.6f,0.6f,0.6f,
		0.6f,0.6f,0.6f,0.6f,0.6f,0.6f,0.6f,
		0.6f,0.6f,0.6f,0.6f,0.6f,0.6f,0.6f
	};

	//normal speeds for each level
	protected float[] level_speeds = new float[21]{
		0.75f,0.85f,0.85f,0.85f,0.95f,0.95f,0.95f,
		0.95f,0.95f,0.95f,0.95f,0.95f,0.95f,0.95f,
		0.95f,0.95f,0.95f,0.95f,0.95f,0.95f,0.95f
	};

	//speeds when moving through the tunnel for each level
	protected float[] tunnel_speeds = new float[21]{
		0.4f,0.45f,0.45f,0.45f,0.5f,0.5f,0.5f,
		0.5f,0.5f,0.5f,0.5f,0.5f,0.5f,0.5f,
		0.5f,0.5f,0.5f,0.5f,0.5f,0.5f,0.5f
	};
	
	//The tile the ghost is attempting to reach
	//in SCATTER mode this is the home corner
	//in CHASE mode each ghost will determine their own target positin relative to pacman
	protected Vector2 target;

	public Ghost(){

	}
	
	//overloaded constructor for making static ghosts on menu; don't want to do a full init()
	public Ghost(bool menu_static) : base() {
	}

	//make only need a single right facing frame for the intro screen
	protected void menuStatic(){
		mesh = GameObject.CreatePrimitive(PrimitiveType.Quad);
		mesh.name = name;
		mesh.transform.localScale = new Vector3(2,-2,1);
		Texture2D tex = Painter(pixelData, true, new int[]{1,2,10,12,13,14,16,17,18}, new int[]{6,8,20}, new int[]{4,5,7,9,11,15,19}, primaryColor, eyeBlue);
		mesh.renderer.material.SetTexture("_MainTex", tex);
		base.init();
	}

	//create all the required animation frames and store them based on mode and facing
	protected override void init(){

		frames.Add(GhostMode.CHASE, new Dictionary<Dir, Texture2D[]>());

		//build the individual Texture2Ds. suppling int values for the body, pupils and whites of the eyes
		frames[GhostMode.CHASE].Add(Dir.LEFT, new Texture2D[]{
			Painter(pixelData, false, new int[]{1,2,10,12,13,14,16,17,18}, new int[]{6,8,20}, new int[]{4,5,7,9,11,15,19}, primaryColor, eyeBlue),
			Painter(pixelData, false, new int[]{1,3,10,12,13,14,16,17,18}, new int[]{6,8,20}, new int[]{4,5,7,9,11,15,19}, primaryColor, eyeBlue)
		});
		
		//RIGHT is just a reverse of the LEFT data
		frames[GhostMode.CHASE].Add(Dir.RIGHT, new Texture2D[]{
			Painter(pixelData, true, new int[]{1,2,10,12,13,14,16,17,18}, new int[]{6,8,20}, new int[]{4,5,7,9,11,15,19}, primaryColor, eyeBlue),
			Painter(pixelData, true, new int[]{1,3,10,12,13,14,16,17,18}, new int[]{6,8,20}, new int[]{4,5,7,9,11,15,19}, primaryColor, eyeBlue)
		});
		frames[GhostMode.CHASE].Add(Dir.UP, new Texture2D[]{
			Painter(pixelData, false, new int[]{1,2,4,6,7,8,10,11,12,17,18,20}, new int[]{14}, new int[]{5,9,13,15,16,19}, primaryColor, eyeBlue),
			Painter(pixelData, false, new int[]{1,3,4,6,7,8,10,11,12,17,18,20}, new int[]{14}, new int[]{5,9,13,15,16,19}, primaryColor, eyeBlue)
		});
		frames[GhostMode.CHASE].Add(Dir.DOWN, new Texture2D[]{
			Painter(pixelData, false, new int[]{1,2,5,6,7,13,14,16,18,19}, new int[]{11,12}, new int[]{4,8,9,10,15,17,20}, primaryColor, eyeBlue),
			Painter(pixelData, false, new int[]{1,3,5,6,7,13,14,16,18,19}, new int[]{11,12}, new int[]{4,8,9,10,15,17,20}, primaryColor, eyeBlue)
		});

		frames.Add(GhostMode.FRIGHTENED, new Dictionary<Dir, Texture2D[]>());

		Texture2D ft0 = Painter(pixelData, false, new int[]{1,2,4,5,6,7,8,9,10,11,12,13,14}, new int[]{15,16,17,18,19,20}, null, frightBlue, frightEye);
		Texture2D ft1 = Painter(pixelData, false, new int[]{1,3,4,5,6,7,8,9,10,11,12,13,14}, new int[]{15,16,17,18,19,20}, null, frightBlue, frightEye);

		//no direction on FRIGHTENED, so just use left
		frames[GhostMode.FRIGHTENED].Add(Dir.LEFT, new Texture2D[]{ft0, ft1});

		//just eyes, no body
		frames.Add(GhostMode.EYES1, new Dictionary<Dir, Texture2D[]>());
		frames[GhostMode.EYES1].Add(Dir.LEFT, new Texture2D[]{
			Painter(pixelData, false, new int[]{}, new int[]{6,8,20}, new int[]{4,5,7,9,11,15,19}, primaryColor, eyeBlue)
		});
		frames[GhostMode.EYES1].Add(Dir.RIGHT, new Texture2D[]{
			Painter(pixelData, true, new int[]{}, new int[]{6,8,20}, new int[]{4,5,7,9,11,15,19}, primaryColor, eyeBlue)
		});
		frames[GhostMode.EYES1].Add(Dir.UP, new Texture2D[]{
			Painter(pixelData, false, new int[]{}, new int[]{14}, new int[]{5,9,13,15,16,19}, primaryColor, eyeBlue)
		});
		frames[GhostMode.EYES1].Add(Dir.DOWN, new Texture2D[]{
			Painter(pixelData, false, new int[]{}, new int[]{11,12}, new int[]{4,8,9,10,15,17,20}, primaryColor, eyeBlue)
		});

		//create a unity quad mesh to hold the texture.
		//textures applied to the material in updateAnimFrame()
		mesh = GameObject.CreatePrimitive(PrimitiveType.Quad);
		mesh.name = name;
		mesh.transform.localScale = new Vector3(2,-2,1);

		base.init();
	}

	public void startLevel(){
		firsPositions();
	}

	//PLACES EVERYONE! PLACES!
	public override void firsPositions ()
	{
		base.firsPositions ();
		enabled = true;
		mode = start_mode;
		doReverse = false; //prevent direction change due to mode change when starting/reset level
		_dir = move_dir = next_move_dir = start_dir;
		last_speed = speed = level_speeds[GameManager.Instance.levelIndex]; //get the speed for this level
		mesh.transform.position = new Vector3(startingPos.x, startingPos.y, -6);
		updateAnimFrame(); //set the texture for this mode/dir
		_halt = false;
		next_tile = new Vector2((int)startingPos.x, -(int)startingPos.y);
	}

	//create a texture and paint each pixel a s[ecofoc colour based on the above map
	protected Texture2D Painter(int[,] data, bool flip, int[] body, int[] pupils, int[] whites, Color bodyColor, Color eyeColor){
		Texture2D tex = new Texture2D(data.GetLength(1),data.GetLength(0));
		tex.filterMode = FilterMode.Point; //sharp pixels
		tex.SetPixels32(blankStamp); //transparency requires Color32 for alpha value, but can only flood fill Color32s
		tex.Apply(); //let the paint dry

		//loop through each entry in the data map, then see if its value is in any of the array for body, pupils and whites. if so draw that pixel. rinse, repeat
		for(int y = 0; y < data.GetLength(1); y++){
			for(int x = 0; x < data.GetLength(0); x++){
				int c = data[y,x];
				if(c > 0){
					Color col = Color.black;
					bool done = false;
					foreach(int b in body){
						if(c == b){ 
							col = bodyColor;
							done = true;
							break;
						}
					}
					if(!done){ //if we haven't already found a body value, check the pupils
						foreach(int p in pupils){
							if(c == p){ 
								col = eyeColor;
								done = true;
								break;
							}
						}
					}
					if(!done){ //no body or pupils, check the whites
						if(whites != null){
							foreach(int w in whites){
								if(c == w){ 
									col = eyeWhite;
									done = true;
									break;
								}
							}
						}
					}

					if(done){ //if found a vlue, paint it, otherwise this pixel remains it's original transparent
						tex.SetPixel((flip? 15-x : x ),y,col);
					}
				}
			}
		}
		tex.Apply();
		return tex;
	}

	//this is a mess. just a horrible mess
	public override void update(float deltaTime){
		if(!_halt){ //pause for when pacman has eaten a ghost. could probably have been handle with the pause in GameManager
			if(_mode != GhostMode.NONE){
				if(_mode == GhostMode.GHOSTHOUSE){ //normal mode within the ghost house. move up and down within the bounds of one tile
					if(move_dir == Dir.DOWN) {
						y += deltaTime * maxSpeed * house_speed;
						if(y > -startingPos.y+0.5f){
							move_dir = Dir.UP;
							_dir = move_dir;
						}
					} else if(move_dir ==  Dir.UP){
						y -= deltaTime * maxSpeed * house_speed;
						if(y < -startingPos.y - 0.5f){
							move_dir = Dir.DOWN;
							_dir = move_dir;
						}
					}
				} else if(_mode == GhostMode.LEAVING){ 
					//trying to get out of the ghost house.
					//first make sure you are on the horizontal centreline of the house
					if(houseStage == 1) {
						if(move_dir == Dir.DOWN) {
							y += deltaTime * maxSpeed * exit_speed;
							if(y > 14.5f){
								y = 14.5f;
								houseStage = 2;
							}
						} else if(move_dir ==  Dir.UP){
							y -= deltaTime * maxSpeed * exit_speed;
							if(y < 14.5){
								y = 14.5f;
								houseStage = 2;
							}
						}
					} else if(houseStage == 2){
						//Inky and Clyde are off the to side, so they need to get in to the middle. Pinky always skips this step
						if(move_dir == Dir.RIGHT){
							//inky
							x += deltaTime * maxSpeed * exit_speed;
							if(x > 14f){
								x = 14f;
								houseStage = 3;
							}
						} else if(move_dir == Dir.LEFT){
							//clyde
							x -= deltaTime * maxSpeed * exit_speed;
							if(x < 14f){
								x = 14f;
								houseStage  = 3;
							}
						}
					} else if(houseStage == 3){
						//finally in position to move out
						y -= deltaTime * maxSpeed * exit_speed;
						if(y < 11.5f){
							y = 11.5f;
							tile_x = (int)x;
							tile_y = (int)y;
							//finally gotten out. get ready to move
							_dir = move_dir = Dir.LEFT;
							//a temporary state, normally only lasting within this frame, but being frightened and exiting have some issues, so needed this
							//this is where ghost house related stuff should've been broken away from the main modes
							mode = GhostMode.OUTSIDE; 
							if(GameManager.Instance.CurrentGhostMode == GhostMode.FRIGHTENED){
								if(was_eyes){ //don't reemerge in frightened mode
									was_eyes = false;
									mode = GameManager.Instance.LastGhostMode;
								} else { //stay frightened, but for realz this time, not the mask of fright been wearing up to this point while in the house
									mode = GameManager.Instance.LastGhostMode;
									mode = GhostMode.FRIGHTENED;
								}
							} else {
								mode = GameManager.Instance.CurrentGhostMode;
							}
							pathfind(); //start moving normally (target based pathfinding movement)
						}
					}
				} else if(_mode == GhostMode.EYES2){
					//this is more or less the same as exiting the house only in reverse to get back in (can't enter the house under normal movement rules
					if(houseStage == 1) {
						if(move_dir == Dir.RIGHT){
							x += deltaTime * maxSpeed * eyes_speed;
							if(x > 14f){
								x = 14f;
								houseStage = 2;
							}
						} else if(move_dir == Dir.LEFT){
							x -= deltaTime * maxSpeed * eyes_speed;
							if(x < 14f){
								x = 14f;
								houseStage = 2;
							}
						}
					} else if(houseStage == 2){
						if(move_dir == Dir.DOWN) {
							y += deltaTime * maxSpeed * eyes_speed;
							if(y > 14.5f){
								y = 14.5f;
								houseStage = 3;
							}
						} else if(move_dir ==  Dir.UP){
							y -= deltaTime * maxSpeed * eyes_speed;
							if(y < 14.5){
								y = 14.5f;
								houseStage = 3;
							}
						}
					} else if(houseStage == 3){
						bool home = false;
						if(move_dir == Dir.RIGHT){
							//clyde
							x += deltaTime * maxSpeed * eyes_speed;
							if(x > startingPos.x){
								x = startingPos.x;
								home = true;
							}
						} else if(move_dir == Dir.LEFT){
							//inky
							x -= deltaTime * maxSpeed * eyes_speed;
							if(x < startingPos.x){
								x = startingPos.x;
								home = true;
							}
						}
						if(home){ 
							//eyes made it back in. now get your 'body' back and go seek your revenge!
							mode = GhostMode.GHOSTHOUSE;
							LeaveGhostHouse();
						}
					}
				} else {
					//finally after all those special movement rules, this is normal movement
					//All in all ghosts are pretty stupid, they only seem smart! 
					//While they have a destination/target in mind they only know how to get to the next tile, and how far that tiles adjacent tiles are from their target.
					//keep moving in this direction. when we get to the centre of the tile check for the next move.
					//we already the next tile we are going to move into (or which direction we need to go to be more specific) but we need to know what to do when we get there
					//so update direction then do some pathfinding!
					float _speed = _mode == GhostMode.EYES1 ? eyes_speed : (tile_id >= 58 ? tunnel_speeds[GameManager.Instance.levelIndex] : speed); //fast when eyes, slow in the tunnel, otherwise keep on trucking
					switch(move_dir){
						case Dir.RIGHT :
							x += deltaTime * maxSpeed * _speed;
							if(x > next_tile.x+0.5f){
								updateDirection();
								pathfind();
							}
							break;
						case Dir.DOWN :
							y += deltaTime * maxSpeed * _speed;
							if(y > next_tile.y+0.5f){
								updateDirection();
								pathfind();
							}
							break;
						case Dir.UP :
							y -= deltaTime * maxSpeed * _speed;
							if(y < next_tile.y+0.5f){
								updateDirection();
								pathfind();
							}
							break;
						case Dir.LEFT :
						default :
							x -= deltaTime * maxSpeed * _speed;
							if(x < next_tile.x+0.5f){
								updateDirection();
								pathfind();
							}
							break;
					}
					
					//just some debug lines in editor to show next tile (can't really see it) and current target.
					Vector3 pn = new Vector3(next_tile.x+0.5f, -next_tile.y-0.5f, -4);
					Vector3 pt = new Vector3(target.x, -target.y, -4);
					Debug.DrawLine(mesh.transform.position, pn, Color.green );
					Debug.DrawLine(mesh.transform.position, pt, Color.yellow );
				}
			}
			if(tile_x == GameManager.Instance.pacman.tileX && tile_y == GameManager.Instance.pacman.tileY){
				//we've moved in to the same tile as pacman. Ooohhh... something is going to happen!!
				if(mode != GhostMode.EYES1 && mode != GhostMode.EYES2){ //you can't eat eyes, silly pacman.
					if(_mode == GhostMode.FRIGHTENED){
						GameManager.Instance.ghostEaten(this); //I die
					} else {
						GameManager.Instance.pacmanEaten(this); //He dies!
					}
				}
			}
			if(mode == GhostMode.EYES1){
				if((int)x == target.x && (int)y == target.y){
					//we've reached the target just above the ghost house. EYES2 is the mode that will get us inside (can't pathfind in to there)
					mode = GhostMode.EYES2;
				}
			}
		}
		base.update(deltaTime);
	}

	public void LeaveGhostHouse(){
		mode = GhostMode.LEAVING;
	}

	//this is a bloody mess! See all previous moaning about modes
	//what needs to happen when ghost changes mode
	public GhostMode mode {
		get { return _mode; }
		set {
			if(value != _mode){
				if(this is Inky){
					//Debug.Log(Time.time+" "+_mode+" "+value+" "+anim_mode_override+" "+houseStage);
				}
				if(mode != GhostMode.EYES1 && mode != GhostMode.EYES2){ //eyes can't change mode until they get to the ghost house
					if(value != GhostMode.GHOSTHOUSE && value != GhostMode.LEAVING && value != GhostMode.OUTSIDE){ //want to change to a 'normal' mode
						if(_mode == GhostMode.GHOSTHOUSE || _mode == GhostMode.LEAVING || _mode == GhostMode.OUTSIDE){ //we are in one of the 'special' modes
							if(value == GhostMode.FRIGHTENED){
								//we want to be frightened, but we can't actually change to that mode as we need to track out situation relative to the house
								//so put on a mask that maske up look frightened. we can't actually be eaten, but pacman can't get to us anyway.
								anim_mode_override = GhostMode.FRIGHTENED;
							} else if( anim_mode_override == GhostMode.FRIGHTENED){ //time to take off the mask
								anim_mode_override = GhostMode.NONE;
								if(_mode == GhostMode.OUTSIDE){
									//FRIGHTENED mode occurred when we were in the middle of the exiting process. How rude!
									//find what out actual speed should be now out in the real world
									_mode = value;
									last_speed = level_speeds[GameManager.Instance.levelIndex];
								}
							} else {
								if(_mode == GhostMode.OUTSIDE){ //this is just normal exiting.
									_mode = value;
									last_speed = level_speeds[GameManager.Instance.levelIndex];
								}
							}
							setModeSpeed(value, anim_mode_override); //if we're wearing the mask we also need to move at the same slower speed as the really frightened ghosts
						} else {
							//this is a 'normal' mode change. so we need to set the reverse flag, so that we know to backtrack on the next direction change
							doReverse = true;
							setModeSpeed(value, _mode);
							_mode = value;
						}
					} else {
						if(value == GhostMode.LEAVING){
							_mode = value;
							houseStage = 1; //exiting the house. start the leaving sequence
						} else if( value == GhostMode.OUTSIDE ){
							_mode = value;
						} else if(value == GhostMode.GHOSTHOUSE){ //entered (or started in) the ghost house. let the GameManager know for the exiting rules
							GameManager.Instance.GhostInDaHouse(this);
							_mode = value;
							house_stage = 0;
						}
					}
				} else {
					_mode = value;
					if(value == GhostMode.EYES2){
						houseStage = 1; //eyes have gotten back to the house. begin docking procedures
						was_eyes = true;
					}
				}
				if(this is Inky){
					//Debug.Log(_mode+" "+speed);
				}
				if(_mode == GhostMode.EYES1 || _mode == GhostMode.EYES2)
					max_frames = 1; //eyes only have one animation frame in each direction.
				else
					max_frames = 2;
			}
		}
	}

	//this manages entering and exiting the ghost house. this just defines what to do next after finishinf a stage. actualy movement is in update
	protected int houseStage {
		get { return house_stage; }
		set { 
			int m = value;
			if(mode == GhostMode.LEAVING){
				if(m == 1){
					if( y < 14.5f){
						_dir = move_dir = Dir.DOWN;
						house_stage = 1;
					} else if( y > 14.5f) {
						_dir = move_dir = Dir.UP;
						house_stage = 1;
					} else {
						m = 2;
					}
				}
				if(m == 2){ //Inky and Clyde only
					if( x < 14f){
						_dir = move_dir = Dir.RIGHT;
						house_stage = 2;
					} else if( x > 14f) {
						_dir = move_dir = Dir.LEFT;
						house_stage = 2;
					} else {
						m = 3;
					}
				}
				if(m == 3){ //exit
					_dir = move_dir = Dir.UP;
					house_stage = 3;
				}
			} else if (mode == GhostMode.EYES2){
				if(m == 1){ //Line up with the gate
					if( x < 14f){
						_dir = move_dir = Dir.RIGHT;
						house_stage = 1;
					} else if( x > 14f) {
						_dir = move_dir = Dir.LEFT;
						house_stage = 1;
					} else {
						m = 2;
					}
				}
				if(m == 2){ //enter the house
					if( y < 14.5f){
						_dir = move_dir = Dir.DOWN;
						house_stage = 2;
					} else if( y > 14.5f) {
						_dir = move_dir = Dir.UP;
						house_stage = 2;
					} else {
						m = 3;
					}
				}
				if(m == 3){ //Inky and Clyde only. find your bed
					if(startingPos.x != 14f){
						if(startingPos.x < 14f){
							_dir = move_dir = Dir.LEFT;
							house_stage = 3;
						} else if( startingPos.x > 14f){
							_dir = move_dir = Dir.RIGHT;
							house_stage = 3;
						}
					} else {
						mode = GhostMode.GHOSTHOUSE;
						LeaveGhostHouse();
					}
				}
			}
		}
	}

	//set and restore spees when entering/exiting frightened mode (or the appearance of frightened mode)
	protected void setModeSpeed(GhostMode new_mode, GhostMode old_mode){
		if(new_mode == GhostMode.FRIGHTENED){
			if(mode == GhostMode.CHASE || mode == GhostMode.SCATTER)
				last_speed = speed;

			speed = fright_speeds[GameManager.Instance.levelIndex];
		} else if( old_mode == GhostMode.FRIGHTENED){
			speed = last_speed;
		}
	}

	//we've hit the center of a tile, set our next, previously determined, direction.
	//or reverse direction if a mode change happened. This is the only time a ghost can backtrack
	override protected void updateDirection(){
		if(doReverse || next_move_dir != move_dir){
			if(mode == GhostMode.CHASE || mode == GhostMode.SCATTER || mode == GhostMode.FRIGHTENED){
				if(doReverse){ //in the arcade version this actually happens then the ghost enters a tile, not getting to the middle
					switch(move_dir){
						case Dir.LEFT: next_move_dir = Dir.RIGHT; break;
						case Dir.RIGHT: next_move_dir = Dir.LEFT; break;
						case Dir.UP: next_move_dir = Dir.DOWN; break;
						case Dir.DOWN: next_move_dir = Dir.UP; break;
					}
					doReverse = false;
				}
			}
			//make sure we're centered
			x = next_tile.x+0.5f;
			y = next_tile.y+0.5f;
			_dir = move_dir = next_move_dir;
		}
	}

	//switch the displayed sprite.
	//this just basically makes sure attempting to extract a valid entry in the Dictionaries. 
	override protected void updateAnimFrame(){
		GhostMode anim_mode = _mode;
		if(anim_mode_override != GhostMode.NONE) {
			anim_mode = anim_mode_override;
		} else {
			if(_mode == GhostMode.SCATTER || _mode == GhostMode.GHOSTHOUSE || _mode == GhostMode.NONE || _mode == GhostMode.LEAVING || _mode == GhostMode.OUTSIDE){
				anim_mode = GhostMode.CHASE;
			}
		}
		if(anim_mode == GhostMode.FRIGHTENED){
			_dir = Dir.LEFT;
		}
		if(anim_mode == GhostMode.EYES2){
			anim_mode = GhostMode.EYES1;
		}
		try {
			if(frames.ContainsKey(anim_mode))
				if(anim_frame >= frames[anim_mode][_dir].Length)
					anim_frame = frames[anim_mode][_dir].Length;

			mesh.renderer.material.SetTexture("_MainTex", frames[anim_mode][_dir][anim_frame]);
		}catch (Exception e)
		{
			Debug.Log(anim_mode+" "+_dir+" "+anim_frame);
			Debug.LogException(e, mesh);
		}
	}

	protected void pathfind(){
		findTarget();
		Vector2 next_ = new Vector2(tile_x, tile_y);
		if(move_dir == Dir.LEFT){
			next_.x--;
		} else if(move_dir == Dir.UP) {
			next_.y--;
		} else if(move_dir == Dir.DOWN){
			next_.y++;
		} else {
			next_.x++;
		}
		//we know this next tile is valid, as we checked in the last path find. but we need to know where to go after this
		next_tile = new Vector2(next_.x, next_.y);
		evalRoute(next_tile);
	}

	protected void evalRoute(Vector2 next_){
		/*
		we want to check each non-wall adjacent tile of the next tile in our current direction.
		we will check each of those tiles and find their distance (as the crow flies) to our current target
		if this distance is shorted that our current shortest distance then we assume that tile is the best choice from that point on
		*/
		Map map = GameManager.Instance.map;
		Vector2 next_maybe = new Vector2();
		Dir next_dir = move_dir;
		float target_range = Mathf.Infinity; //want to find the smallest value, so start with a really big value

		int nt = map.getTileId(next_.x, next_.y);
		if(nt == 0 || nt > 49) { //sanity check empty. should never get here is this was false, but you never know
			//wall values are 1 to 33 (the number of Tile types in map). just assume anything above 50 is safe to move on (include dots, tunnels etc)
			if(move_dir != Dir.LEFT){ //no backtracking
				int ntr = map.getTileId(next_.x+1, next_.y);//check right
				if(ntr == 0 || ntr > 49){ //this tile is valid
					float dr = Vector2.Distance(target, new Vector2(next_.x+1, next_.y));
					if(dr <= target_range){ //is this smaller than our current smallest distance
						//yes, so this might be the chosen one
						target_range = dr; 
						next_maybe = next_;
						next_maybe.x++;
						next_dir = Dir.RIGHT;
					}
				}
			}
			//keep going for all valid directions
			if(move_dir != Dir.UP){
				int ntd = map.getTileId(next_.x, next_.y+1);//check down
				if(ntd == 0 || ntd > 49 ){
					float dd = Vector2.Distance(target, new Vector2(next_.x, next_.y+1));
					if(dd <= target_range){
						target_range = dd;
						next_maybe = next_;
						next_maybe.y++;
						next_dir = Dir.DOWN;
					}
				}
			}
			if(move_dir != Dir.RIGHT){
				int ntl = map.getTileId(next_.x-1, next_.y);//check left
				if(ntl == 0 || ntl > 49){
					float dl = Vector2.Distance(target, new Vector2(next_.x-1, next_.y));
					if(dl <= target_range){
						target_range = dl;
						next_maybe = next_;
						next_maybe.x--;
						next_dir = Dir.LEFT;
					}
				}
			}
			if(move_dir != Dir.DOWN){
				int ntu = map.getTileId(next_.x, next_.y-1);//check up
				if(!(ntu > 0 && ntu < 50) && !(_mode != GhostMode.FRIGHTENED && (nt == 54 || nt == 56))){ //can't go up beside where Blinky starts (in front of the ghost house) or beside where Pac-Man starts
					float du = Vector2.Distance(target, new Vector2(next_.x, next_.y-1));
					if(du <= target_range){ //order of preference on equal ranges, up, left, down, right. hence this order of if statements
						target_range = du;
						next_maybe = next_;
						next_maybe.y--;
						next_dir = Dir.UP;
					}
				}
			}
			next_move_dir = next_dir;
		}
	}

	//pick a target assuming sub classes haven't already
	protected virtual void findTarget(){
		if(_mode == GhostMode.EYES1) {
			target = house_target; //disembodied eyes have only one place to go
		} else {
			//this applies to FRIGHTENED mode too, which it shouldn't.
			//FRIGHTENED mode uses a pseudo-random nubmer generator to decide where to go at each intersection (i.e there is no target). if initial direction blocked, check clockwise
			//not implimented due to time.
			target = homePos;
		}
	}

	//Pinky, Inky & Clyde override this to determine their exit conditions.
	public virtual int dotsLimit() {
		return 0;
	}

	//used by Pinky, Inky & Clyde to determine when to leave the ghost hose. Blinky never stays in the ghost house and only enters after been eaten, then leaves right away
	//see GameManager.addGhostDot()
	public bool dotEaten(){
		if(mode == GhostMode.GHOSTHOUSE){
			if(++dot_counter >= dotsLimit()){
				LeaveGhostHouse();
				return true;
			}
		}
		return false;
	}

	public Color color {
		get { return primaryColor; }
	}
}
